/*
 * Author: Xu Yifeng
 * Email : skypexu@gmail.com
 * Date  : 2009/06/26
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include <errno.h>

#include "erl_driver.h"

/* op codes */
#define ID_OPEN		0
#define ID_CONV		1
#define ID_CLOSE	2
#define ID_RESET	3

/* conversion buffer size */
#define OUTBUF_SIZE	4096

#define ASIZE(spec)	(sizeof(spec)/sizeof(spec[0]))

static ErlDrvEntry	iconvdrv_entry;

/* atoms which are sent to erlang VM */
static ErlDrvTermData	am_ok;
static ErlDrvTermData	am_error;
static ErlDrvTermData	am_enomem;
static ErlDrvTermData	am_einval;
static ErlDrvTermData	am_eilseq;
static ErlDrvTermData	am_unknown;
static ErlDrvTermData	am_noent;

typedef struct iconvdrv_t {
	ErlDrvPort	port;
	ErlDrvTermData	port_id;
	char		outbuf[OUTBUF_SIZE];
	iconv_t		handle;
} iconvdrv_t;

static ErlDrvData
iconvdrv_start(ErlDrvPort port, char *buf)
{
	iconvdrv_t *icd = (iconvdrv_t *)driver_alloc(sizeof(iconvdrv_t));

	if (icd == NULL)
		return (ErlDrvData)-1;
	icd->port = port;
	icd->port_id = driver_mk_port(port);
	icd->handle = (iconv_t)-1;
	driver_pdl_create(icd->port);
	return (ErlDrvData) icd;
}

static void
iconvdrv_stop(ErlDrvData drv_data)
{
	iconvdrv_t *icd = (iconvdrv_t *)drv_data;

	if (icd->handle != (iconv_t)-1)
		iconv_close(icd->handle);
	driver_free(icd);
}

static int
driver_send_data(iconvdrv_t *icd, ErlDrvBinary *bin, int len)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, icd->port_id,
		ERL_DRV_ATOM, am_ok,
		ERL_DRV_BINARY, (ErlDrvTermData)bin, len, 0,
		ERL_DRV_TUPLE, 3};
	ErlDrvTermData to = driver_caller(icd->port);
	return driver_send_term(icd->port, to, spec, ASIZE(spec));
}

static int
driver_send_ok(iconvdrv_t *icd)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, icd->port_id,
		ERL_DRV_ATOM, am_ok,
		ERL_DRV_TUPLE, 2};
	ErlDrvTermData to = driver_caller(icd->port);
	return driver_send_term(icd->port, to, spec, ASIZE(spec));
}

static int
driver_send_error(iconvdrv_t *icd, ErlDrvTermData am)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, icd->port_id,
		ERL_DRV_ATOM, am_error,
		ERL_DRV_ATOM, am,
		ERL_DRV_TUPLE, 3};
	ErlDrvTermData to = driver_caller(icd->port);
	return driver_send_term(icd->port, to, spec, ASIZE(spec));
}

static int
driver_send_error2(iconvdrv_t *icd, ErlDrvTermData am, int pos, ErlDrvBinary *bin, int len)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, icd->port_id,
		ERL_DRV_ATOM, am_error,
		ERL_DRV_ATOM, am,
		ERL_DRV_INT, pos,
		ERL_DRV_BINARY, (ErlDrvTermData)bin, len, 0,
		ERL_DRV_TUPLE, 5};
	ErlDrvTermData to = driver_caller(icd->port);
	return driver_send_term(icd->port, to, spec, ASIZE(spec));
}

static int
get_int32(char **buf, char *end, int *val)
{
	if (*buf + 4 > end)
		return (-1);

	*val = ((unsigned char *)(*buf))[0] << 24 | \
	       ((unsigned char *)(*buf))[1] << 16 | \
	       ((unsigned char *)(*buf))[2] << 8  | \
	       ((unsigned char *)(*buf))[3];
	*buf += 4;
	return (0);
}

static int
get_string(char **buf, char *end, char **string)
{
	int len;

	*string = NULL;
	if (get_int32(buf, end, &len) == -1)
		return (-1);
	if (end - *buf < len)
		return (-1);
	*string = strndup(*buf, len);
	*buf += len;
	return (0);
}

static void
i_open(iconvdrv_t *icd, char *tocode, char *fromcode)
{
	iconv_t	h;

	if ((h = iconv_open(tocode, fromcode)) == (iconv_t)-1) {
		driver_send_error(icd, am_einval);
	} else {
		icd->handle = h;
		driver_send_ok(icd);
	}
	return;
}

static void
i_conv(iconvdrv_t *icd, const char *ip, size_t ileft)
{
	const char *orig = ip;
	size_t	oleft;
	size_t	ret;
	char	*op;
	int	len;
	int	totlen;
	ErlDrvBinary	*bin;
	iconv_t	h;

	if ((h = icd->handle) == (iconv_t)-1) {
		driver_send_error(icd, am_noent);
		return;
	}
	bin    = NULL;
	totlen = 0;
retry:
	oleft = OUTBUF_SIZE;
	op = &icd->outbuf[0];
	ret = iconv(h, &ip, &ileft, &op, &oleft);
	len = OUTBUF_SIZE - oleft;
	if (bin == NULL) {
		if ((bin = driver_alloc_binary(len)) == NULL) {
			driver_send_error(icd, am_enomem);
			return;
		}
	} else {
		ErlDrvBinary *new_bin;
		new_bin = driver_realloc_binary(bin, totlen + len);
		if (new_bin == NULL) {
			driver_free_binary(bin);
			driver_send_error(icd, am_enomem);
			return;
		}
		bin = new_bin;
	}
	memcpy(bin->orig_bytes+totlen, &icd->outbuf[0], len);
	totlen += len;

	if (ret == (size_t)-1) {
		if (errno == EILSEQ)
			driver_send_error2(icd, am_eilseq, ip-orig, bin, totlen);
		else if (errno == EINVAL)
			driver_send_error2(icd, am_einval, ip-orig, bin, totlen);
		else if (errno == E2BIG)
			goto retry;
		else 
			driver_send_error2(icd, am_unknown, ip-orig, bin, totlen);
	} else {
		driver_send_data(icd, bin, totlen);
	}
	driver_free_binary(bin);
	return;
}

static void
i_close(iconvdrv_t *icd)
{
	if (icd->handle != (iconv_t)-1) {
		iconv_close(icd->handle);
		icd->handle = (iconv_t)-1;
	}
	driver_send_ok(icd);
	return;
}

static void
iconvdrv_from_erlang(ErlDrvData drv_data, char *buf, int len)
{
	iconvdrv_t *icd = (iconvdrv_t *)drv_data;
	char *tocode, *fromcode;
	char *end, *bp = buf;
	unsigned int i;

	i = bp[0];
	end = bp + len;
	bp++;

	switch (i) {
	case ID_OPEN:
		tocode = NULL;
		fromcode = NULL;
		if (get_string(&bp, end, &tocode) == -1 ||
		    get_string(&bp, end, &fromcode) == -1) {
			driver_send_error(icd, am_enomem);
		} else {
			i_open(icd, tocode, fromcode);
		}
		free(tocode);
		free(fromcode);
		break;

	case ID_CONV:
		{
		int n;
		if (get_int32(&bp, end, &n) == -1)
			driver_send_error(icd, am_enomem);
		else
			i_conv(icd, bp, n);
		}
		break;

	case ID_CLOSE:
		i_close(icd);
		break;

	case ID_RESET:
		if (icd->handle != (iconv_t)-1) {
			iconv(icd->handle, NULL, NULL, NULL, NULL);
			driver_send_ok(icd);
		} else {
			driver_send_error(icd, am_noent);
		}
	}
	return;
}

DRIVER_INIT(iconvdrv)
{

	am_ok		= driver_mk_atom("ok");
	am_error	= driver_mk_atom("error");
	am_enomem	= driver_mk_atom("enomem");
	am_einval	= driver_mk_atom("einval");
	am_eilseq	= driver_mk_atom("eilseq");
	am_unknown	= driver_mk_atom("unknown");
	am_noent	= driver_mk_atom("enoent");

	iconvdrv_entry.init         = NULL;
	iconvdrv_entry.start        = iconvdrv_start;
	iconvdrv_entry.stop         = iconvdrv_stop;
	iconvdrv_entry.output       = iconvdrv_from_erlang;
	iconvdrv_entry.ready_input  = NULL;
	iconvdrv_entry.ready_output = NULL;
	iconvdrv_entry.driver_name  = "iconv_drv";
	iconvdrv_entry.finish       = NULL;
	iconvdrv_entry.outputv      = NULL;
	iconvdrv_entry.extended_marker = ERL_DRV_EXTENDED_MARKER;
	iconvdrv_entry.major_version = ERL_DRV_EXTENDED_MAJOR_VERSION;
	iconvdrv_entry.minor_version = ERL_DRV_EXTENDED_MINOR_VERSION;
	iconvdrv_entry.driver_flags = ERL_DRV_FLAG_USE_PORT_LOCKING;
	return &iconvdrv_entry;
}
